/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf.layers;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonTypeInfo;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.*;
import lombok.*;
import lombok.experimental.SuperBuilder;
import lombok.extern.slf4j.Slf4j;
import net.brutex.ai.dnn.api.ILayerConfiguration;
import net.brutex.ai.dnn.api.LayerType;
import org.deeplearning4j.nn.api.ParamInitializer;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.dropout.Dropout;
import org.deeplearning4j.nn.conf.dropout.IDropout;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.weightnoise.IWeightNoise;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.regularization.Regularization;

/** A neural network layer. */
@JsonTypeInfo(use = JsonTypeInfo.Id.CLASS, include = JsonTypeInfo.As.PROPERTY, property = "@class")
@EqualsAndHashCode
// @JsonIdentityInfo(generator= ObjectIdGenerators.IntSequenceGenerator.class, property="@id")
@Slf4j
@SuperBuilder
public abstract class LayerConfiguration
    implements ILayerConfiguration, Serializable, Cloneable { // ITrainableLayerConfiguration

  @Getter @Setter protected String name;
  @Getter protected List<LayerConstraint> allParamConstraints;
  @Getter protected List<LayerConstraint> weightConstraints;
  @Getter protected List<LayerConstraint> biasConstraints;
  @Getter protected List<LayerConstraint> constraints;
  @Getter @Setter protected IWeightNoise weightNoise;
  @Builder.Default private @Getter @Setter LinkedHashSet<String> variables = new LinkedHashSet<>();
  @Getter @Setter private IDropout dropOut;
  /** The type of the layer, basically defines the base class and its properties */
  @Builder.Default @Getter @Setter @NonNull private LayerType type = LayerType.UNKNOWN;
  /**
   * A reference to the neural net configuration. This field is excluded from json serialization as
   * well as from equals check to avoid circular referenced.
   *
   * <p>-- GETTER -- Get the neural net configuration this layer configuration is part of
   *
   * @return neural net configuration
   *     <p>-- SETTER -- Set the neural net configuration for this layer
   * @param neural net configuration
   */
  @Getter @Setter @JsonIgnore @EqualsAndHashCode.Exclude
  private NeuralNetConfiguration netConfiguration;
  /**
   * From an Activation, we can derive the IActivation (function) using {@link
   * Activation#getActivationFunction()} but not vice versa. The default is Identity Activation.
   */
  @Builder.Default @Getter @Setter private IActivation activation = Activation.IDENTITY;

  /**
   * Get the activation interface (function) from the activation. The activation must have been set
   * prior to this call.
   *
   * @return activation function
   */
  public IActivation getActivationFn() {
    if (activation == null)
      throw new RuntimeException(
          "No activation is set for this Layer. Cannot return activation function.");
    return activation;
  }

  public void addVariable(String s) {
    variables.add(s);
  }

  public String toJson() {
    throw new RuntimeException("toJson is not implemented for LayerConfiguration");
  }

  /** Initialize the weight constraints. Should be called last, in the outer-most constructor */
  protected void initializeConstraints() {
    // Note: this has to be done AFTER all constructors have finished - otherwise the required
    // fields may not yet be set yet
    List<LayerConstraint> allConstraints = new ArrayList<>();
    if (allParamConstraints != null && !initializer().paramKeys(this).isEmpty()) {
      for (LayerConstraint c : allParamConstraints) {
        LayerConstraint c2 = c.clone();
        c2.setParams(new HashSet<>(initializer().paramKeys(this)));
        allConstraints.add(c2);
      }
    }

    if (weightConstraints != null && !initializer().weightKeys(this).isEmpty()) {
      for (LayerConstraint c : weightConstraints) {
        LayerConstraint c2 = c.clone();
        c2.setParams(new HashSet<>(initializer().weightKeys(this)));
        allConstraints.add(c2);
      }
    }

    if (biasConstraints != null && !initializer().biasKeys(this).isEmpty()) {
      for (LayerConstraint c : biasConstraints) {
        LayerConstraint c2 = c.clone();
        c2.setParams(new HashSet<>(initializer().biasKeys(this)));
        allConstraints.add(c2);
      }
    }
    if (!allConstraints.isEmpty()) {
      this.constraints = allConstraints;
    } else {
      this.constraints = null;
    }
  }

  /**
   * Reset the learning related configs of the layer to default. When instantiated with a global
   * neural network configuration the parameters specified in the neural network configuration will
   * be used. For internal use with the transfer learning API. Users should not have to call this
   * method directly.
   */
  public void resetLayerDefaultConfig() {
    // clear the learning related params for all layers in the origConf and set to defaults
    this.dropOut = null;
    this.constraints = null;
  }

  /**
   * Migration workaround //TODO To be removed
   *
   * @return a layer configuration
   */
  @Deprecated
  @JsonIgnore
  public LayerConfiguration getLayer() {
    return this;
  }

  @Override
  public LayerConfiguration clone() {
    try {
      LayerConfiguration ret = (LayerConfiguration) super.clone();
      // Let's check for any INDArray fields and dup them (in case cloned layer will be used in
      // different threads on CUDA...
      // we don't want it being relocated contantly between devices)
      Class<?> c = getClass();
      while (c != Object.class) {
        Field[] fields = c.getDeclaredFields();
        for (Field f : fields) {
          if (f.getType() == INDArray.class) {
            f.setAccessible(true);
            INDArray toClone;
            try {
              toClone = (INDArray) f.get(this);
            } catch (Exception e) {
              throw new RuntimeException(e);
            }
            if (toClone != null) {
              try {
                f.set(this, toClone.dup());
              } catch (Exception e) {
                throw new RuntimeException(e);
              }
            }
          }
        }

        c = c.getSuperclass();
      }

      return ret;
    } catch (CloneNotSupportedException e) {
      throw new RuntimeException(e);
    }
  }

  public abstract org.deeplearning4j.nn.api.Layer instantiate(
      @NonNull NeuralNetConfiguration conf,
      Collection<TrainingListener> trainingListeners,
      int layerIndex,
      INDArray layerParamsView,
      boolean initializeParams,
      DataType networkDataType);

  /**
   * @return The parameter initializer for this model
   */
  public abstract ParamInitializer initializer();

  /**
   * For a given type of input to this layer, what is the type of the output?
   *
   * @param layerIndex Index of the layer
   * @param inputType Type of input for the layer
   * @return Type of output from the layer
   * @throws IllegalStateException if input type is invalid for this layer
   */
  public abstract InputType getOutputType(int layerIndex, InputType inputType);

  /**
   * Set the nIn value (number of inputs, or input channels for CNNs) based on the given input type
   *
   * @param inputType Input type for this layer
   * @param override If false: only set the nIn value if it's not already set. If true: set it
   *     regardless of whether it's already set or not.
   * @throws IllegalStateException if input type is invalid for this layer
   */
  public abstract void setNIn(InputType inputType, boolean override);

  /**
   * For the given type of input to this layer, what preprocessor (if any) is required?<br>
   * Returns null if no preprocessor is required, otherwise returns an appropriate {@link
   * InputPreProcessor} for this layer, such as a {@link
   * org.deeplearning4j.nn.conf.preprocessor.CnnToFeedForwardPreProcessor}
   *
   * @param inputType InputType to this layer
   * @return Null if no preprocessor is required, otherwise the type of preprocessor necessary for
   *     this layer/input combination
   * @throws IllegalStateException if input type is invalid for this layer
   */
  public abstract InputPreProcessor getPreProcessorForInputType(InputType inputType);

  /**
   * Get the regularization types (l1/l2/weight decay) for the given parameter. Different parameters
   * may have different regularization types.
   *
   * @param paramName Parameter name ("W", "b" etc)
   * @return Regularization types (if any) for the specified parameter
   */
  public abstract List<Regularization> getRegularizationByParam(String paramName);

  /**
   * Is the specified parameter a layerwise pretraining only parameter?<br>
   * For example, visible bias params in an autoencoder (or, decoder params in a variational
   * autoencoder) aren't used during supervised backprop.<br>
   * Layers (like DenseLayerConfiguration, etc) with no pretrainable parameters will return false
   * for all (valid) inputs.
   *
   * @param paramName Parameter name/key
   * @return True if the parameter is for layerwise pretraining only, false otherwise
   */
  public abstract boolean isPretrainParam(String paramName);

  /**
   * Get the updater for the given parameter. Typically, the same updater will be used for all
   * parameters, but this is not necessarily the case
   *
   * @param paramName Parameter name
   * @return IUpdater for the parameter
   */
  public IUpdater getUpdaterByParam(String paramName) {
    throw new UnsupportedOperationException(
        "Not supported: all layers with parameters should override this method");
  }

  public IUpdater getIUpdater() {
    throw new UnsupportedOperationException(
        "Not supported: all layers with parameters should override this method");
  }

  public void setIUpdater(IUpdater iUpdater) {
    log.warn(
        "Setting an IUpdater on {} with name {} has no effect.", getClass().getSimpleName(), name);
  }

  /**
   * This is a report of the estimated memory consumption for the given layer
   *
   * @param inputType Input type to the layer. Memory consumption is often a function of the input
   *     type
   * @return Memory report for the layer
   */
  public abstract LayerMemoryReport getMemoryReport(InputType inputType);

  public void clearVariables() {
    this.variables.clear();
  }

  /**
   * Inherit setting from neural network for those settings, that are not already set or do have a
   * layer(type) specific default. This implementation does not require the neural network
   * configuration to be the same as the one returned from this layers {@link
   * #getNetConfiguration()}.
   *
   * @param conf a neural net configration to inherit parameters from
   */
  public void runInheritance(@NonNull NeuralNetConfiguration conf) {
    if (this.activation == null) this.activation = conf.getActivation();
    if (this.dropOut == null) this.dropOut = conf.getIdropOut();
    if (this.weightNoise == null) this.weightNoise = conf.getWeightNoise();
  }

  /**
   * Runs {@link #runInheritance(NeuralNetConfiguration)} using the layers configurations embedded
   * neural net configuration (the one returned from {@link #getNetConfiguration()}.
   */
  public void runInheritance() {
    runInheritance(getNetConfiguration());
  }

  public abstract static class LayerConfigurationBuilder<
      C extends LayerConfiguration, B extends LayerConfigurationBuilder<C, B>> {
    private String name;
    private List<LayerConstraint> allParamConstraints;
    private List<LayerConstraint> weightConstraints;
    private List<LayerConstraint> biasConstraints;
    private List<LayerConstraint> constraints;
    private IWeightNoise weightNoise;
    private LinkedHashSet<String> variables$value;
    private boolean variables$set;
    private IDropout dropOut;
    private @NonNull LayerType type$value;
    private boolean type$set;
    private NeuralNetConfiguration netConfiguration;
    private IActivation activation$value;
    private boolean activation$set;

    public B activation(Activation activation) {
      this.activation$value = activation;
      this.activation$set = true;
      return self();
    }

    public B activation(IActivation activation) {
      this.activation$value = activation;
      this.activation$set = true;
      return self();
    }

    public B dropOut(double d) {
      this.dropOut = new Dropout(d);
      return self();
    }

    public B dropOut(IDropout d) {
      this.dropOut = d;
      return self();
    }

    public B constrainBias(LayerConstraint constraint) {
      return this.biasConstraints(List.of(constraint));
    }

    public B constrainWeights(LayerConstraint constraint) {
      return this.weightConstraints(List.of(constraint));
    }

    public B name(String name) {
      this.name = name;
      return self();
    }

    public B allParamConstraints(List<LayerConstraint> allParamConstraints) {
      this.allParamConstraints = allParamConstraints;
      return self();
    }

    public B weightConstraints(List<LayerConstraint> weightConstraints) {
      this.weightConstraints = weightConstraints;
      return self();
    }

    public B biasConstraints(List<LayerConstraint> biasConstraints) {
      this.biasConstraints = biasConstraints;
      return self();
    }

    public B constraints(List<LayerConstraint> constraints) {
      this.constraints = constraints;
      return self();
    }

    public B weightNoise(IWeightNoise weightNoise) {
      this.weightNoise = weightNoise;
      return self();
    }

    public B variables(LinkedHashSet<String> variables) {
      this.variables$value = variables;
      this.variables$set = true;
      return self();
    }

    public B type(@NonNull LayerType type) {
      this.type$value = type;
      this.type$set = true;
      return self();
    }

    @JsonIgnore
    public B netConfiguration(NeuralNetConfiguration netConfiguration) {
      this.netConfiguration = netConfiguration;
      return self();
    }

    protected abstract B self();

    public abstract C build();

    public String toString() {
      return "LayerConfiguration.LayerConfigurationBuilder(name="
          + this.name
          + ", allParamConstraints="
          + this.allParamConstraints
          + ", weightConstraints="
          + this.weightConstraints
          + ", biasConstraints="
          + this.biasConstraints
          + ", constraints="
          + this.constraints
          + ", weightNoise="
          + this.weightNoise
          + ", variables$value="
          + this.variables$value
          + ", variables$set="
          + this.variables$set
          + ", dropOut="
          + this.dropOut
          + ", type$value="
          + this.type$value
          + ", type$set="
          + this.type$set
          + ", netConfiguration="
          + this.netConfiguration
          + ", activation$value="
          + this.activation$value
          + ", activation$set="
          + this.activation$set
          + ", variables$value="
          + this.variables$value
          + ", type$value="
          + this.type$value
          + ", activation$value="
          + this.activation$value
          + ")";
    }
  }
}
